/* **************************************************************************



Copyright (c) 2000 Novell, Inc.  All Rights Reserved.



THIS WORK IS  SUBJECT  TO  U.S.  AND  INTERNATIONAL  COPYRIGHT  LAWS  AND

TREATIES.   NO  PART  OF  THIS  WORK MAY BE  USED,  PRACTICED,  PERFORMED

COPIED, DISTRIBUTED, REVISED, MODIFIED, TRANSLATED,  ABRIDGED, CONDENSED,

EXPANDED,  COLLECTED,  COMPILED,  LINKED,  RECAST, TRANSFORMED OR ADAPTED

WITHOUT THE PRIOR WRITTEN CONSENT OF NOVELL, INC. ANY USE OR EXPLOITATION

OF THIS WORK WITHOUT AUTHORIZATION COULD SUBJECT THE PERPETRATOR TO

CRIMINAL AND CIVIL LIABILITY.



****************************************************************************/

using System;

using DXMLBase;

using System.Xml;

using System.Threading;



namespace DXMLNullDriver

{



	/// <summary>Identity Manager Null Driver Shim.

	/// </summary>

	/// <version>3.6

	/// </version>

    public class NullDriverShim : IDriverShimV1, IPublicationShim, IXmlQueryProcessor, ISubscriptionShim, ISubscriptionServiceShim

	{

		private const string DRIVER_ID_VALUE = "NULLSVC";

		private static readonly string DRIVER_VERSION_VALUE = getVersionFromManifest();

		private const string DRIVER_MIN_ACTIVATION_VERSION = "0";



		/// <summary> Variable used in <code>start()</code> to determine when to return from <code>start()</code>.</summary>

		private bool shutdownFlag = false;



		/// <summary> Object used as a gate so subscriber-channel thread can wake publisher thread

		/// up to tell it to shutdown.

		/// </summary>

		internal Object shutdownGate = new Object();



		/// <summary> A Java driver shim must have a constructor which takes no parameters. DirXML

		/// uses this to construct the driver object.

		/// </summary>

		public NullDriverShim()

		{

		}



		/// <summary> <code>init</code> will be called after a driver is created to allow the driver

		/// to perform any necessary initialization before starting to process events and

		/// commands. Typically it will perform any setup that is common to both the

		/// subscriber and publisher channels.

		/// </summary>

		/// <param name="initParameters">XML doc that contains the shim initialization	parameters

		/// </param>

		/// <returns> XML doc containing status/messages from the init operation

		/// </returns>

		public virtual XmlDocument init(XmlDocument initParameters)

		{

			setDriverRDN(initParameters);

			if (heartbeatInterval == 0)

			{

				heartbeatInterval = getHeartbeatInterval(initParameters);

				if (heartbeatInterval < 0)

				{

					heartbeatInterval = 0;

					return createStatusDocument(STATUS_WARNING, "Ignoring invalid value for <pub-heartbeat-interval>");

				}

			}

			return createSuccessDocument();

		}



		/// <summary> <code>shutdownShim</code> indicates to the DriverShim that

		/// the driver is being terminated.

		/// 

		/// </summary>

		/// <param name="reason">XML doc that may contain description of reason for shutdown

		/// </param>

		/// <returns> XML doc containing status/messages from shutdown operation

		/// </returns>

		public virtual XmlDocument shutdown(XmlDocument reason)

		{

			try

			{

				lock (shutdownGate)

				{

					shutdownFlag = true;

					Monitor.Pulse(shutdownGate);

				}

				return createSuccessDocument();

			}

			catch (Exception t)

			{

				//something bad happened...

				return createStatusDocument(STATUS_FATAL, t.Message);

			}

		}



		/// <summary> <code>getSubscriptionShim</code> gets an implementation of SubscriptionShim

		/// that can be used to process commands on behalf of DirXML.

		/// </summary>

		/// <returns> an instance of  SubscriptionShim that can be used to process

		/// commands on behalf of the DirXML.

		/// </returns>

		public virtual ISubscriptionShim getSubscriptionShim()

		{

			return this;

		}

        public virtual ISubscriptionServiceShim getSubscriptionServiceShim()
        {

            return this;

        }


		/// <summary> <code>getPublicationShim</code> gets an implementation of PublicationShim

		/// that can be invoked by NDS.

		/// </summary>

		/// <returns> an instance of PublicationShim that can be invoked by DirXML.

		/// </returns>

		public virtual IPublicationShim getPublicationShim()

		{

			return this;

		}



		/// <summary> <code>getSchema</code> returns a representation of the application schema data encoded in XML.

		/// <p>

		/// This will be called only when the driver is not running. In other words, if this method is

		/// called init()/shutdown() will not be called for the current instance of the DriverShim.

		/// </summary>

		/// <param name="initParameters">XML doc containing the driver shim initialization parameters as

		/// well as the subscription shim and publication shim initialization parameters.

		/// </param>

		/// <returns> XML doc containing the application schema data, or an error status return.

		/// </returns>

		public virtual XmlDocument getSchema(XmlDocument initParameters)

		{

			//Return an empty schema

			setDriverRDN(initParameters);

			XmlElement output = createOutputDocument();

			XmlDocument doc = output.OwnerDocument;

			XmlElement schemadef = doc.CreateElement("schema-def");

			output.AppendChild(schemadef);

			addStatusElement(output, STATUS_SUCCESS, null, null);

			return output.OwnerDocument;

		}



		/// <summary> <code>execute</code> will execute a command encoded in an XML doc.

		/// 

		/// </summary>

		/// <param name="doc	The">doc that contains the commands

		/// </param>

		/// <param name="query	A">query processor that can be used to query the the caller for more

		/// information

		/// </param>

		/// <returns> This function will return a doc that contains information

		/// of changes that should occur in NDS as a result of the application

		/// processing that changes that were passed in on this call. For example,

		/// the object was created in the application space and here is its

		/// unique identifier to add to the NDS object's association... It may also

		/// contain status and logging information.

		/// </returns>

		public virtual XmlDocument execute(XmlDocument doc, IXmlQueryProcessor query)

		{

			try

			{

				XmlElement outputElement = createOutputDocument();

				XmlDocument outputDoc = outputElement.OwnerDocument;

				XmlElement inputElement = (XmlElement)doc.GetElementsByTagName("input", String.Empty).Item(0);

				if (inputElement != null)

				{

					foreach (XmlNode node in inputElement.ChildNodes)

					{

						if (node.NodeType == XmlNodeType.Element)

						{

							XmlElement operation = (XmlElement)node;

							if (operation.NamespaceURI.Length == 0)

							{

								// handle driver identification query

								if (operation.LocalName.Equals("query") && (operation.SelectSingleNode("search-class[@class-name='__driver_identitification_class__']") != null))

								{

									addDriverIdentification(outputElement, query);

								}

								else if (operation.LocalName.Equals("query-schema"))

								{

									// handle schema query

									XmlElement schemadef = outputDoc.CreateElement("schema-def");

									String eventID = operation.GetAttribute("event-id");

									if (eventID.Length > 0)

									{

										schemadef.SetAttribute("event-id", eventID);

									}

									outputElement.AppendChild(schemadef);

								}

							}

							// every operation should return a success status

							addStatusElement(outputElement, STATUS_SUCCESS, null, operation.GetAttribute("event-id"));

						}

					}

				}

				return outputElement.OwnerDocument;

			}

			catch (Exception t)

			{

				return createStatusDocument(STATUS_ERROR, t.Message);

			}

		}



		/// <summary> Add driver identification information to the output doc in response to

		/// an engine query.

		/// 

		/// </summary>

		/// <param name="output">output element of output doc.

		/// </param>

		internal virtual void addDriverIdentification(XmlElement output, IXmlQueryProcessor query)

		{

			XmlDocument doc = output.OwnerDocument;

			XmlElement instance = doc.CreateElement("instance");

			output.AppendChild(instance);

			instance.SetAttribute("class-name", "__driver_identification_class__");

			XmlElement attr = doc.CreateElement("attr");

			instance.AppendChild(attr);

			attr.SetAttribute("attr-name", "driver-id");

			XmlElement val = doc.CreateElement("value");

			attr.AppendChild(val);

			val.SetAttribute("type", "string");

			XmlText text = doc.CreateTextNode(DRIVER_ID_VALUE);

			val.AppendChild(text);

			attr = doc.CreateElement("attr");

			instance.AppendChild(attr);

			attr.SetAttribute("attr-name", "driver-version");

			val = doc.CreateElement("value");

			attr.AppendChild(val);

			val.SetAttribute("type", "string");

			text = doc.CreateTextNode(DRIVER_VERSION_VALUE);

			val.AppendChild(text);

			attr = doc.CreateElement("attr");

			instance.AppendChild(attr);

			attr.SetAttribute("attr-name", "min-activation-version");

			val = doc.CreateElement("value");

			attr.AppendChild(val);

			val.SetAttribute("type", "int");

			text = doc.CreateTextNode(DRIVER_MIN_ACTIVATION_VERSION);

			val.AppendChild(text);

			attr = doc.CreateElement("attr");

			instance.AppendChild(attr);

			attr.SetAttribute("attr-name", "query-ex-supported");

			val = doc.CreateElement("value");

			attr.AppendChild(val);

			val.SetAttribute("type", "state");

			text = doc.CreateTextNode("false");

			val.AppendChild(text);

		}





		// implement the PublicationShim interface

		public virtual XmlDocument start(IXmlCommandProcessor executePublisher)

		{

			long nextHeartbeat = (DateTime.Now.Ticks - 621355968000000000) / 10000 + heartbeatInterval;

			lock (shutdownGate)

			{

				while (!shutdownFlag)

				{

					try

					{

						if (heartbeatInterval != 0)

						{

							long currentTime = (DateTime.Now.Ticks - 621355968000000000) / 10000;

							if (currentTime < nextHeartbeat)

							{

								Monitor.Wait(shutdownGate, TimeSpan.FromMilliseconds(nextHeartbeat - currentTime));

							}

							nextHeartbeat = (DateTime.Now.Ticks - 621355968000000000) / 10000 + heartbeatInterval;



							// send the heartbeat

							XmlElement input = createInputDocument();

							XmlElement status = addStatusElement(input, STATUS_SUCCESS, null, null);

							status.SetAttribute("type", "heartbeat");

							executePublisher.execute(input.OwnerDocument, this);

						}

						else

						{

							Monitor.Wait(shutdownGate);

						}

					}

					catch (ThreadInterruptedException)

					{

					}

				}

			}

			return createSuccessDocument();

		}



		/// <summary> <code>query</code> will accept an XML-encoded query and

		/// return the results

		/// 

		/// </summary>

		/// <param name="doc	A">doc containing an XDS encoded query

		/// </param>

		/// <returns> The results of the query

		/// </returns>

		public virtual XmlDocument query(XmlDocument doc)

		{

			return createSuccessDocument();

		}





		/// <summary> Create a return status doc for DirXML that signals success

		/// 

		/// </summary>

		/// <returns>	An XmlDocument with a status element, level="success"

		/// </returns>

		private XmlDocument createSuccessDocument()

		{

			return createStatusDocument(STATUS_SUCCESS, null);

		}



		/// <summary> Create an XDS status doc for returning status to DirXML.

		/// 

		/// </summary>

		/// <param name="level">STATUS_SUCCESS, STATUS_FATAL, STATUS_ERROR, STATUS_WARNING, or STATUS_RETRY

		/// </param>

		/// <param name="message">A detail message that will be the content of the status element (may be null)

		/// </param>

		private XmlDocument createStatusDocument(int level, string message)

		{

			//create a doc for returning something to DirXML

			XmlElement output = createOutputDocument();

			//add the status element

			addStatusElement(output, level, message, null);

			return output.OwnerDocument;

		}



		/// <summary> Add a &lt;status> element to an XDS input or output doc. The doc must already

		/// have an &lt;output> element or an &lt;input> element.

		/// 

		/// </summary>

		/// <param name="parent">The input or output element in the XDS doc.

		/// </param>

		/// <param name="level">STATUS_SUCCESS, STATUS_FATAL, STATUS_ERROR, STATUS_WARNING, or STATUS_RETRY

		/// </param>

		/// <param name="message">A detail message that will be the content of the status element (may be null)

		/// </param>

		/// <param name="eventId">The event id to which the status element corresponds (may be null)

		/// </param>

		/// <returns> The added status element.

		/// </returns>

		private XmlElement addStatusElement(XmlElement parent, int level, string message, string eventId)

		{

			//get the DOM Document for use as a factory

			XmlDocument document = parent.OwnerDocument;

			//create the status element and place it under the parent element

			XmlElement status = document.CreateElement("status");

			parent.AppendChild(status);

			//set the level based on what was passed

			status.SetAttribute("level", STATUS_LEVELS[level]);

			//if we have an event id, set the attribute

			if (eventId != null && eventId.Length > 0)

			{

				status.SetAttribute("event-id", eventId);

			}

			//if there is a detail message, put in in as the content of the status element

			if (message != null && message.Length > 0)

			{

				XmlText msg = document.CreateTextNode(message);

				status.AppendChild(msg);

			}

			return status;

		}



		/// <summary> Create a bare-bones XDS doc for DirXML use, containing only the root

		/// &lt;nds> element.

		/// 

		/// </summary>

		/// <returns>	The root nds Element

		/// </returns>

		private XmlElement createXdsDocument()

		{

			XmlDocument document = new XmlDocument();

			//create "nds" doc element

			XmlElement nds = document.CreateElement("nds");

			document.AppendChild(nds);

			nds.SetAttribute("dtdversion", "3.5");

			XmlElement source = document.CreateElement("source");

			nds.AppendChild(source);

			XmlElement product = document.CreateElement("product");

			source.AppendChild(product);

			product.AppendChild(document.CreateTextNode("DirXML Null Driver"));

			//		product.setAttributeNS(null, "build",Build.getBuildString());

			product.SetAttribute("instance", driverRDN);

			product.SetAttribute("version", DRIVER_VERSION_VALUE);

			XmlElement contact = document.CreateElement("contact");

			source.AppendChild(contact);

			contact.AppendChild(document.CreateTextNode("Novell, Inc."));

			return nds;

		}



		/// <summary> Create an empty "output" doc for returning something to DirXML. Return the

		/// output element.

		/// 

		/// </summary>

		/// <returns> the output <code>Element</code>.

		/// </returns>

		private XmlElement createOutputDocument()

		{

			//create the basic XDS doc

			XmlElement nds = createXdsDocument();

			//add an output element

			XmlElement output = nds.OwnerDocument.CreateElement("output");

			nds.AppendChild(output);

			return output;

		}



		/// <summary> Create an empty "input" doc for submitting something (publishing or subscriber query) to DirXML. Return the

		/// input element.

		/// 

		/// </summary>

		/// <returns> the input <code>Element</code>.

		/// </returns>

		private XmlElement createInputDocument()

		{

			//create the basic XDS doc

			XmlElement nds = createXdsDocument();

			//add an input element

			XmlElement input = nds.OwnerDocument.CreateElement("input");

			nds.AppendChild(input);

			return input;

		}



		/// <summary>Get the RDN from a DN

		/// </summary>

		/// <param name="dn">the DN

		/// </param>

		/// <returns> the RDN

		/// </returns>

		static private string getRDN(string dn)

		{

			string rdn = dn;

			int index = rdn.LastIndexOf("\\");

			if ((index != -1) && (index < rdn.Length))

			{

				rdn = rdn.Substring(index + 1);

			}

			return rdn;

		}



		/// <summary> Get the heartbeat value, if any, from the init doc

		/// </summary>

		/// <param name="initDoc">The doc passed to init()

		/// </param>

		/// <returns> The value from the doc in milliseconds or 0, if none found, or -1 if invalid

		/// </returns>

		private int getHeartbeatInterval(XmlDocument initDoc)

		{

			XmlElement optionsElement = (XmlElement)initDoc.GetElementsByTagName("publisher-options", String.Empty).Item(0);

			if (optionsElement == null)

			{

				return 0;

			}

			XmlElement heartbeatElement = (XmlElement)optionsElement.GetElementsByTagName("pub-heartbeat-interval", String.Empty).Item(0);

			if (heartbeatElement == null)

			{

				return 0;

			}

			string val = heartbeatElement.InnerText;

			long interval;

			try

			{

				interval = Int64.Parse(val);

			}

			catch (FormatException)

			{

				return -1;

			}

			//interval is specified in options in minutes, but we want it internally as milliseconds

			interval *= 60 * 1000;

			if (interval > Int32.MaxValue)

			{

				interval = -1;

			}

			return (int)interval;

		}



		/// <summary>Set the driver rdn from an init doc.

		/// </summary>

		/// <param name="doc">the init doc.

		/// </param>

		private void setDriverRDN(XmlDocument initDoc)

		{

			if (driverRDN.Length == 0)

			{

				//find <init-params> element

				XmlElement initParams = (XmlElement)initDoc.GetElementsByTagName("init-params", String.Empty).Item(0);

				if (initParams != null)

				{

					//return the driver DN

					string dn = initParams.GetAttribute("src-dn");

					driverRDN = getRDN(dn);

				}

			}



		}

		/// <summary> Get the version string from the jar file manifest</summary>

		/// <returns> the version string from the manifest

		/// </returns>

		private static string getVersionFromManifest()

		{

			// FIXIT - I'm sure there is a way to do this

			string version = "?.?.?.?";

			//try

			//{

			//    Type.GetType("com.novell.nds.dirxml.driver.loopback.LoopbackDriverShim");

			//    Uri url = new Uri(System.IO.Path.GetFullPath("LoopbackDriverShim.class"));

			//    url = new Uri(url, "../../../../../../META-INF/MANIFEST.MF");

			//    Manifest mf = new Manifest(System.Net.WebRequest.Create(url).GetResponse().GetResponseStream());

			//    Attributes atts = mf.getMainAttributes();

			//    string val = atts.getValue(Attributes.Name.IMPLEMENTATION_VERSION);

			//    if ((val != null) && (val.Length != 0))

			//    {

			//        version = val;

			//    }

			//}

			//catch (Exception t)

			//{

			//}

			return version;



		}





		private string driverRDN = "";

		private int heartbeatInterval = 0;



		/// <summary>	Parameter to status doc creation methods for level attribute of "success".</summary>

		private const int STATUS_SUCCESS = 0;

		/// <summary>	Parameter to status doc creation methods for level attribute of "fatal".</summary>

		private const int STATUS_FATAL = 1;

		/// <summary>	Parameter to status doc creation methods for level attribute of "error".</summary>

		private const int STATUS_ERROR = 2;

		/// <summary>	Parameter to status doc creation methods for level attribute of "warning".</summary>

		private const int STATUS_WARNING = 3;

		/// <summary>	Parameter to status doc creation methods for level attribute of "retry".</summary>

		private const int STATUS_RETRY = 4;



		/// <summary> Array used to translate int level parameters to XML attribute values</summary>

		private static readonly string[] STATUS_LEVELS = new string[] { "success", "fatal", "error", "warning", "retry" };

	}

}
